;;;
;;;	EMILY Board  (Shared Memory Platform 4K)
;;;

	include	"config.inc"

;;; AVR local definitions
XL	reg	r26
XH	reg	r27
YL	reg	r28
YH	reg	r29
ZL	reg	r30
ZH	reg	r31

	include	"../common.inc"
	include	"../emily.inc"
	
PB_CS0	equ	0
PB_CS1	equ	1
PB_CS2	equ	2
PB_CS3	equ	3
PB_RES	equ	4

PD_ALEH	equ	4
PD_ALEL	equ	5
PD_RD	equ	6
PD_WR	equ	7
	
	
;;;
;;; ROM area
;;;
	
	segment	code

	org	0x0000
	rjmp	cstart

	;; org	0x000E
	;; rjmp	pci3_int
	
	org	0x0020
	rjmp	t0_int

	org	0x0100

cstart:
	cli
	;; Initialize Stack Pointer
	ldi	r16,low(STACK)
	out	SPL,r16
	ldi	r16,high(STACK)
	out	SPH,r16

	rcall	init

	;; Initialize ports
	ldi	r16,0b11111111
	out	PORTA,r16
	clr	r16
	out	DDRA,r16

	ldi	r16,0b00001111
	out	PORTB,r16
	ldi	r16,0b00011111
	out	DDRB,r16

	ldi	r16,0b11111100
	out	PORTC,r16

	ldi	r16,0b11110000
	out	PORTD,r16
	out	DDRD,r16
	
	;; Initialize UART1
        ldi     r16,0		; 9600bps (19.2MHz)
        sts     UBRR1H,r16
        ldi     r16,124		; 9600bps (19.2MHz)
        sts     UBRR1L,r16
        ldi     r16,0b00000000
        sts     UCSR1A,r16
        ldi     r16,0b00011000	; Enable Tx,Rx
        sts     UCSR1B,r16
        ldi     r16,0b00000110	; 8bit NONE
        sts     UCSR1C,r16

	;; Initialize timer 0
	ldi	r16,0b00000010	; CTC
	out	TCCR0A,r16
	ldi	r16,0b00001100	; CTC, /256
	out	TCCR0B,r16
	ldi	r16,14		; 200us
	out	OCR0A,r16
	ldi	r16,0b00000010
	sts	TIMSK0,r16

	;; Initialize TWI (I2C)
	ldi	r16,22		; TWBR=22 (100kHz)
	sts	TWBR,r16
	ldi	r16,1		; TWPS=1
	sts	TWSR,r16
	
        ;; Initialize work area
        clr     r1
        sts     saddrh,r1
        sts     saddrl,r1
        sts     dsadrh,r1
        sts     dsadrl,r1

	sts	timer,r1	; timer = 0

	ldi	r16,1		; Try target EEPROM
	sts	eenum,r16
	rcall	epread
	brcc	cs0

	sts	eenum,r1	; Try internal EEPROM
	sts	eenum,r16
	rcall	epread
	brcc	cs0

	ldi	r16,0x0f	; 0FF0 - 0FFF
	sts	shmadr,r16
	ldi	r16,0xf0	; 0FF0 - 0FFF
	sts	shmadr+1,r16
	sts	shmoff,r1	; +1
	sts	mmode,r1
        ldi     r16,'I'
        sts     hexmod,r16
cs0:	
	sei
	
	;; Opening message
        ldi     ZL,low(opnmsg)
        ldi     ZH,high(opnmsg)
        rcall   msgout
	rcall	mdetect
	rcall	mddisp
	rcall	mdd
	rcall	add
	
wstart:
        ldi     r16,'>'
        rcall   conout
        ldi     r16,' '
        rcall   conout
        rcall   getlin
        ldi     XL,low(inbuf)
        ldi     XH,high(inbuf)
        rcall   skipsp
        tst     r16
        breq    wstart          ; Null command
        rcall   upper

	cpi	r16,'D'
	brne	ml0
	rjmp	dump
ml0:
	cpi	r16,'S'
	brne	ml1
	rjmp	setm
ml1:
        cpi     r16,'L'
        brne    ml2
	rjmp    loadh
ml2:
	cpi	r16,'P'
	brne	ml3
	rjmp	saveh
ml3:
	cpi	r16,'A'
	brne	ml4
	rjmp	addr
ml4:
	cpi	r16,'T'
	brne	ml5
	rjmp	target
ml5:
	cpi	r16,'H'
	brne	ml6
	rjmp	mode
ml6:
	cpi	r16,'E'
	brne	ml7
	rjmp	eesub
ml7:
	
err:
	ldi	ZL,low(errmsg)
	ldi	ZH,high(errmsg)
	rcall	msgout
	rjmp	wstart

	;;
	;; D(ump)
	;; 
dump:
	adiw	XL,1
	rcall	skipsp
	rcall	rdhex0
	tst	r13
	brne	dp0
	; No arg.
	rcall	skipsp
	tst	r16
	brne	dpe
	lds	r10,dsadrh
	lds	r11,dsadrl
	ldi	r16,128
	add	r11,r16
	adc	r10,r1
	sts	deadrh,r10
	sts	deadrl,r11
	rjmp	dpm
dpe:
	rjmp	err

dp0:
	sts	dsadrh,r14
	sts	dsadrl,r15
	rcall	skipsp
	cpi	r16,','
	breq	dp1
	tst	r16
	brne	dpe
	; No 2nd arg.
	ldi	r16,128
	add	r15,r16
	adc	r14,r1
	sts	deadrh,r14
	sts	deadrl,r15
	rjmp	dpm
dp1:
	adiw	XL,1
	rcall	skipsp
	rcall	rdhex0
	rcall	skipsp
	tst	r13
	breq	dpe
	tst	r16
	brne	dpe
	sec
	adc	r15,r1
	adc	r14,r1
	sts	deadrh,r14
	sts	deadrl,r15
dpm:
	lds	r10,dsadrh
	lds	r11,dsadrl
	ldi	r16,0xf0
	and	r11,r16
	sts	dstate,r1
dpm0:
	rcall	dpl
	rcall	const
	tst	r16
	brne	dpm1
	lds	r16,dstate
	cpi	r16,2
	brcs	dpm0
	lds	r10,deadrh
	lds	r11,deadrl
dpm1:
	sts	dsadrh,r10
	sts	dsadrl,r11
	rjmp	wstart

dpl:
	mov	r16,r10
	rcall	hexout2
	mov	r16,r11
	rcall	hexout2
	ldi	ZH,high(dsep0)
	ldi	ZL,low(dsep0)
	rcall	msgout
	ldi	ZH,high(inbuf)
	ldi	ZL,low(inbuf)
	ldi	r19,16
dpl0:
	rcall	dpb
	dec	r19
	brne	dpl0

	ldi	ZH,high(dsep1)
	ldi	ZL,low(dsep1)
	rcall	msgout

	ldi	ZH,high(inbuf)
	ldi	ZL,low(inbuf)
	ldi	r19,16
dpl1:
	ld	r16,Z+
	cpi	r16,' '
	brcs	dpl2
	cpi	r16,0x7f
	brcc	dpl2
	rcall	conout
	rjmp	dpl3
dpl2:
	ldi	r16,'.'
	rcall	conout
dpl3:
	dec	r19
	brne	dpl1
	rjmp	crlf

dpb:
	ldi	r16,' '
	rcall	conout
	lds	r16,dstate
	tst	r16
	brne	dpb2
	; Dump state 0
	lds	r16,dsadrl
	cp	r16,r11
	brne	dpb0
	lds	r16,dsadrh
	cp	r16,r10
	breq	dpb1
dpb0:
	ldi	r16,' '
	rcall	conout
	rcall	conout
	st	Z+,r16
	sec
	adc	r11,r1
	adc	r10,r1
	ret
dpb1:
	ldi	r16,1
	sts	dstate,r16
dpb2:
	lds	r16,dstate
	cpi	r16,1
	brne	dpb0
	;; rcall	open
	rcall	read
	;; rcall	close
	st	Z+,r8
	mov	r16,r8
	rcall	hexout2
	sec
	adc	r11,r1
	adc	r10,r1
	lds	r16,deadrl
	cp	r16,r11
	brne	dpbr
	lds	r16,deadrh
	cp	r16,r10
	brne	dpbr
	ldi	r16,2
	sts	dstate,r16
dpbr:
	ret

	;;
	;; S(et)
	;; 
setm:
	adiw	XL,1
	rcall	skipsp
	rcall	rdhex0
	rcall	skipsp
	tst	r16
	brne	sete0
	tst	r13
	brne	stm0	; argument found
	lds	r10,saddrh
	lds	r11,saddrl
	rjmp	stm1
sete0:
	rjmp	err
stm0:
	mov	r10,r14
	mov	r11,r15
stm1:
	mov	r16,r10
	rcall	hexout2
	mov	r16,r11
	rcall	hexout2
	ldi	ZL,low(dsep1)
	ldi	ZH,high(dsep1)
	rcall	msgout
	;; rcall	open
	rcall	read
	;; rcall	close
	mov	r16,r8
	rcall	hexout2
	ldi	r16,' '
	rcall	conout
	rcall	getlin
	ldi	XL,low(inbuf)
	ldi	XH,high(inbuf)
	rcall	skipsp
	tst	r16
	brne	stm2
	; Empty
	sec
	adc	r11,r1
	adc	r10,r1
	sts	saddrh,r10
	sts	saddrl,r11
	rjmp	stm1
stm2:
	cpi	r16,'-'
	brne	sm3
	sec
	sbc	r11,r1
	sbc	r10,r1
	sts	saddrh,r10
	sts	saddrl,r11
	rjmp	stm1
sm3:
	cpi	r16,'.'
	brne	sm4
	rjmp	wstart
sm4:
	rcall	rdhex0
	tst	r13
	breq	sete
	mov	r8,r15
	;; rcall	open
	rcall	write
	;; rcall	close
	sec
	adc	r11,r1
	adc	r10,r1
	sts	saddrh,r10
	sts	saddrl,r11
	rjmp	stm1
sete:
	rjmp	err

	;;
	;; L(oad)
	;; 
loadh:
	clr	r16
	sts	leaddr,r16
	sts	leaddr+1,r16
	ldi	r16,0xFF
	sts	lsaddr,r16
	sts	lsaddr+1,r16

	adiw	XL,1
	rcall	skipsp
	rcall	rdhex0
	rcall	skipsp
	tst	r16
	brne	sete
	tst	r13
	brne	lh0

	clr	r14
	clr	r15
lh0:
	rcall	conin
	rcall	upper
	cpi	r16,'S'
	brne	lh1
	rjmp	lhs1
lh1:
	cpi	r16,':'
	brne	lh0

	ldi	XH,high(inbuf)
	ldi	XL,low(inbuf)
	clr	r19
	clr	r20
lhi1:
	rcall	hexin
	ldi	r18,2
	cp	r9,r18
	brne	lhi2
	st	X+,r8
	add	r20,r8
	inc	r19
	cpi	r19,BUFLEN
	brcs	lhi1
lhi2:
	tst	r20
	brne	lhie		; Checksum error
	lds	r20,inbuf	; Length
	subi	r20,-5		; (Len+Adr*2+Typ+Sum)
	cp	r19,r20
	breq	lhi3
lhie:
	ldi	ZH,high(ihemsg)
	ldi	ZL,low(ihemsg)
	rcall	msgout
	rjmp	wstart
lhi3:
	lds	r20,inbuf	; Length
	tst	r20
	breq	lhir
	lds	r19,inbuf+3	; Type
	tst	r19
	breq	lhi4		; Data record
	cpi	r19,1
	breq	lhir		; End record
	rjmp	lh0		; Unknown record
lhi4:
	lds	r10,inbuf+1
	lds	r11,inbuf+2
	add	r11,r15
	adc	r10,r14
	ldi	XH,high(inbuf+4)
	ldi	XL,low(inbuf+4)
	;; rcall	open
lhi5:
	ld	r8,X+
	rcall	lhr
	rcall	write
	sec
	adc	r11,r1
	adc	r10,r1
	dec	r20
	brne	lhi5
	;; rcall	close
	ldi	r16,'I'
	sts	hexmod,r16
	rjmp	lh0
lhir:
	rcall	lhd
	rjmp	wstart

lhs1:
	rcall	conin
	subi	r16,'0'
	brcc	lhs3
lhs2:
	rjmp	lh0     ; Record is not a number
lhs3:
	cpi	r16,10
	brcc	lhs2
	ldi	XH,high(inbuf)
	ldi	XL,low(inbuf)
	st	X+,r16	; Record number
	ldi	r19,1
	clr	r20
lhs4:
	rcall	hexin
	ldi	r18,2
	cp	r9,r18
	brne	lhs5
	st	X+,r8
	add	r20,r8
	inc	r19
	cpi	r19,BUFLEN
	brcs	lhs4
lhs5:
	cpi	r20,0xff
	brne	lhse
	lds	r20,inbuf+1	; Length
	subi	r20,-2		; (Typ+Len)
	cp	r19,r20
	breq	lhs6
lhse:
	ldi	ZH,high(shemsg)
	ldi	ZL,low(shemsg)
	rcall	msgout
	rjmp	wstart
lhs6:
	ldi	XH,high(inbuf)
	ldi	XL,low(inbuf)
	ld	r19,X+		; Type
	ld	r20,X+		; Length
	cpi	r19,1
	brne	lhs61
	; Type 1
	ld	r10,X+
	ld	r11,X+
	subi	r20,2+1		; (Adr*2+Sum)
	rjmp	lhs7
lhs61:
	cpi	r19,2
	brne	lhs62
	; Type 2
	adiw	XL,1		; Skip A23-A16
	ld	r10,X+
	ld	r11,X+
	subi	r20,3+1		; (Adr*3+Sum)
	rjmp	lhs7
lhs62:
	cpi	r19,3
	brne	lhs63
	; Type 3
	adiw	XL,2		; Skip A31-A16
	ld	r10,X+
	ld	r11,X+
	subi	r20,4+1		; (Adr*4+Sum)
	rjmp	lhs7
lhs63:
	cpi	r19,9
	breq	lhsr
	cpi	r19,8
	breq	lhsr
	cpi	r19,7
	breq	lhsr
	rjmp	lh0		; Unknown record
lhs7:
	add	r11,r15
	adc	r10,r14
	;; rcall	open
lhs8:
	ld	r8,X+
	rcall	lhr
	rcall	write
	sec
	adc	r11,r1
	adc	r10,r1
	dec	r20
	brne	lhs8
	;; rcall	close
	ldi	r16,'S'
	sts	hexmod,r16
	rjmp	lh0
lhsr:
	rcall	lhd
	rjmp	wstart

lhr:
	lds	r16,lsaddr
	cp	r10,r16
	breq	lhr0
	brcs	lhr1
	rjmp	lhr2
lhr0:
	lds	r16,lsaddr+1
	cp	r11,r16
	brcc	lhr2
lhr1:
	sts	lsaddr,r10
	sts	lsaddr+1,r11
lhr2:
	lds	r16,leaddr
	cp	r10,r16
	breq	lhr3
	brcc	lhr4
	rjmp	lhr5
lhr3:
	lds	r16,leaddr+1
	cp	r11,r16
	brcs	lhr5
lhr4:
	sts	leaddr,r10
	sts	leaddr+1,r11
lhr5:	
	ret

lhd:
        ldi     ZL,low(ldrmsg)
        ldi     ZH,high(ldrmsg)
        rcall   msgout
	lds	r16,lsaddr
	rcall	hexout2
	lds	r16,lsaddr+1
	rcall	hexout2
	ldi	r16,'-'
	rcall	conout
	lds	r16,leaddr
	rcall	hexout2
	lds	r16,leaddr+1
	rcall	hexout2
	rjmp	crlf

	;;
	;; P(unch)
	;; 
saveh:
	adiw	XL,1
	ld	r16,X
	rcall	upper
	cpi	r16,'I'
	breq	sh0
	cpi	r16,'S'
	brne	sh1
sh0:
	adiw	XL,1
	sts	hexmod,r16
sh1:
	rcall	skipsp
	rcall	rdhex0		; Start address
	tst	r13
	breq	she
	mov	r10,r14
	mov	r11,r15
	rcall	skipsp
	cpi	r16,','
	brne	she
	adiw	XL,1
	rcall	skipsp
	rcall	rdhex0		; End address
	tst	r13
	breq	she
	rcall	skipsp
	tst	r16
	breq	sh2
she:
	rjmp	err
sh2:
	sub	r15,r11
	sbc	r14,r10
	sec
	adc	r15,r1
	adc	r14,r1
sh3:
	rcall	shl
	mov	r16,r14
	or	r16,r15
	brne	sh3	
	lds	r16,hexmod
	cpi	r16,'I'
	brne	sh4
	ldi	ZH,high(ihexer)
	ldi	ZL,low(ihexer)
	rcall	msgout
	rjmp	wstart
sh4:
	ldi	ZH,high(srecer)
	ldi	ZL,low(srecer)
	rcall	msgout
	rjmp	wstart

shl:
	ldi	r19,16
	tst	r14
	brne	shl0
	cp	r15,r19
	brcc	shl0
	mov	r19,r15
shl0:
	lds	r16,hexmod
	cpi	r16,'I'
	brne	shls

	;; Intel HEX
	clr	r20
	ldi	r16,':'
	;sub	r20,r16
	rcall	conout
	mov	r16,r19
	sub	r20,r16
	rcall	hexout2		; Length
	mov	r16,r10
	sub	r20,r16
	rcall	hexout2		; Address (H)
	mov	r16,r11
	sub	r20,r16
	rcall	hexout2		; Address (L)
	clr	r16
	sub	r20,r16
	rcall	hexout2		; Type
	sub	r15,r19
	sbc	r14,r1
	;; rcall	open
shli0:
	rcall	read
	mov	r16,r8
	sub	r20,r8
	rcall	hexout2
	sec
	adc	r11,r1
	adc	r10,r1
	dec	r19
	brne	shli0
	;; rcall	close
	mov	r16,r20
	rcall	hexout2		; Checksum
	rcall	crlf
	ret

shls:
	; Motorola S record
	ldi	r20,0xff
	ldi	r16,'S'
	rcall	conout
	ldi	r16,'1'
	rcall	conout
	mov	r16,r19
	subi	r16,-(2+1)	; (Adr*2+Sum)
	sub	r20,r16
	rcall	hexout2		; Length
	mov	r16,r10
	sub	r20,r16
	rcall	hexout2		; Address (H)
	mov	r16,r11
	sub	r20,r16
	rcall	hexout2		; Address (L)
	sub	r15,r19
	sbc	r14,r1
	;; rcall	open
shls2:
	rcall	read
	mov	r16,r8
	sub	r20,r8
	rcall	hexout2
	sec
	adc	r11,r1
	adc	r10,r1
	dec	r19
	brne	shls2
	;; rcall	close
	mov	r16,r20
	rcall	hexout2		; Checksum
	rcall	crlf
	ret

	;;
	;; A(ddress)
	;; 
addr:
	adiw	XL,1
	rcall	skipsp
	ld	r16,X
	tst	r16
	breq	ad2
	
	rcall	rdhex0
	tst	r13
	breq	ade
	rcall	skipsp

	sts	shmadr,r14
	sts	shmadr+1,r15
	sts	shmoff,r1
	
	ld	r16,X+
	tst	r16
	breq	ad2

	cpi	r16,','
	brne	ade
	rcall	skipsp
	ld	r16,X+
	ld	r13,X
	tst	r13
	brne	ade
	cpi	r16,'1'
	breq	ad2
	cpi	r16,'2'
	brne	ad0
	ldi	r16,1
	sts	shmoff,r16
	rjmp	ad2
ad0:
	cpi	r16,'4'
	brne	ade
	ldi	r16,3
	sts	shmoff,r16
ad2:
	rcall	add
adr:	
	rjmp	wstart
ade:
	rjmp	err

add:	
	ldi	ZH,high(shmmsg0)
	ldi	ZL,low(shmmsg0)
	rcall	msgout
	lds	r16,shmadr
	rcall	hexout2
	lds	r16,shmadr+1
	andi	r16,0b11110011
	rcall	hexout2
	
	ldi	ZH,high(shmmsg1)
	ldi	ZL,low(shmmsg1)
	rcall	msgout

	lds	r16,shmoff
	andi	r16,0b00000011
	inc	r16
	ori	r16,'0'
	rcall	conout

	ldi	r16,')'
	rcall	conout

	rjmp	crlf

	;;
	;; T(arget)
	;; 
target:
	adiw	XL,1
	ld	r16,X
	rcall	upper
	cpi	r16,'R'
	breq	treset
	cpi	r16,'H'
	breq	thalt
te:	
	rjmp	err

treset:
	adiw	XL,1
	ld	r16,X
	tst	r16
	brne	te

	sbi	DDRB,PB_RES
	ldi	r16,50
	sts	timer,r16
tr0:
	lds	r16,timer
	tst	r16
	brne	tr0
	cbi	DDRB,PB_RES

	rjmp	wstart

thalt:
	adiw	XL,1
	ld	r16,X
	tst	r16
	brne	te

	sbi	DDRB,PB_RES

	rjmp	wstart

	;;
	;; (H) Memory Mode
	;;
mode:
	adiw	XL,1
	rcall	skipsp
	ld	r16,X
	tst	r16
	breq	md1
	
	rcall	rdhex0
	tst	r13
	breq	ade
	rcall	skipsp
	tst	r16
	brne	ade
	;;
	sts	mmode,r15
md1:
	rcall	mdd
	rjmp	wstart

mdd:
	ldi	ZH,high(mmdmsg)
	ldi	ZL,low(mmdmsg)
	rcall	msgout
	lds	r16,mmode
	rcall	hexout2
	rjmp	crlf
	
	;;
	;; (E)EPROM sub-commands
	;;
eesub:
	adiw	XL,1
	ld	r16,X+
	rcall	upper
	cpi	r16,'R'
	brne	ees0
	rjmp	epr
ees0:
	cpi	r16,'W'
	brne	ees1
	rjmp	epw
ees1:
	cpi	r16,'S'
	brne	ees2
	rjmp	ems
ees2:
	cpi	r16,'L'
	brne	ees3
	rjmp	eml
ees3:
eese:
	rjmp	err

	;; (ER) EEPROM parameter read
epr:
	rcall	skipsp
	ld	r16,X+
	tst	r16
	breq	epr1
	subi	r16,'0'
	ld	r17,X
	tst	r17
	breq	epr0
	rjmp	err
epr0:
	sts	eenum,r16
epr1:
	rcall	epread
	brcs	eese
	rjmp	wstart	

epread:
	ldi	XL,low(eptmp)
	ldi	XH,high(eptmp)
	clr	YL
	clr	YH
	ldi	r17,32
eprd0:
	rcall	eeread
	brcs	eprde
	st	X+,r8
	adiw	YL,1
	dec	r17
	brne	eprd0

	ldi	XL,low(eptmp)
	ldi	XH,high(eptmp)
	ld	r16,X+
	cpi	r16,0xFF
	brne	eprde
	ld	r16,X+
	cpi	r16,0xAA
	brne	eprde
	ld	r16,X+
	cpi	r16,0x55
	brne	eprde
	ld	r16,X+
	cpi	r16,0x50
	brne	eprde
	
	ldi	XL,low(eptmp)
	ldi	XH,high(eptmp)
	ldi	YL,low(epbuf)
	ldi	YH,high(epbuf)
	ldi	r17,32
eprd1:
	ld	r16,X+
	st	Y+,r16
	dec	r17
	brne	eprd1
	clc
	ret
eprde:
	sec
	ret
	
	;; (EW) EEPROM parameter write
epw:
	rcall	skipsp
	ld	r16,X+
	tst	r16
	breq	epw1
	subi	r16,'0'
	ld	r17,X
	tst	r17
	breq	epw0
	rjmp	err
epw0:
	sts	eenum,r16
epw1:
	rcall	epwrite
	brcs	epwe
	rjmp	wstart	
epwe:
	rjmp	err
	
epwrite:	
	ldi	XL,low(epbuf)
	ldi	XH,high(epbuf)
	ldi	r16,0xFF
	st	X+,r16
	ldi	r16,0xAA
	st	X+,r16
	ldi	r16,0x55
	st	X+,r16
	ldi	r16,0x50
	st	X+,r16

	ldi	XL,low(epbuf)
	ldi	XH,high(epbuf)
	clr	YL
	clr	YH
	ldi	r17,32
epwr0:
	rcall	eeread
	brcs	epwre
	ld	r16,X+
	cp	r16,r8
	breq	epwr1

	mov	r8,r16
	rcall	eewrite
epwr1:
	adiw	YL,1
	dec	r17
	brne	epwr0

	clc
	ret	
epwre:
	sec
	ret

	;; (ES) EEPROM memory save
ems:
	rcall	skipsp
	ld	r16,X
	tst	r16
	breq	ems0

	rcall	rdhex0
	tst	r13
	breq	emse
	sts	emadr,r14
	sts	emadr+1,r15
	rcall	skipsp
	ld	r16,X+
	cpi	r16,','
	brne	emse
	rcall	skipsp
	rcall	rdhex0
	tst	r13
	breq	emse
	rjmp	ems1
ems0:
	lds	r16,lsaddr
	sts	emadr,r16
	lds	r16,lsaddr+1
	sts	emadr+1,r16
	lds	r14,leaddr
	lds	r15,leaddr+1
ems1:
	lds	r16,lsaddr+1
	sub	r15,r16
	lds	r16,lsaddr
	sbc	r14,r16
	sec
	adc	r15,r1
	adc	r14,r1
	sts	emlen,r14
	sts	emlen+1,r15

	rcall	emsb
	rcall	emsm
	rjmp	wstart
	
emse:	
	rjmp	err

emsm:				; Save Meta Data
	ldi	XL,low(emsig0)	; Signature
	ldi	XH,high(emsig0)
	ldi	r16,0xFF
	st	X+,r16
	ldi	r16,0xAA
	st	X+,r16
	ldi	r16,0x55
	st	X+,r16
	ldi	r16,0x53
	st	X+,r16

	ldi	XL,low(embuf)
	ldi	XH,high(embuf)
	ldi	YL,low(128)
	ldi	YH,high(128)
	ldi	r17,16
emsm0:
	ld	r8,X+
	rcall	ee1write
	brcs	emsme
	adiw	YL,1
	dec	r17
	brne	emsm0
	clc
	ret
emsme:
	sec
	ret

emsb:				; Save Binary Data
	lds	XH,emadr
	lds	XL,emadr+1
	ldi	YH,high(256)
	ldi	YL,low(256)
	lds	r19,emlen
	lds	r20,emlen+1
emsb0:
	push	YH
	push	YL
	mov	r10,XH
	mov	r11,XL
	rcall	read
	pop	YL
	pop	YH
	rcall	ee1write
	brcs	emsbe
	adiw	XL,1
	adiw	YL,1
	sec
	sbc	r20,r1
	sbc	r19,r1
	mov	r16,r20
	or	r16,r19
	brne	emsb0
	clc
	ret
emsbe:
	sec
	ret
	
	;; (EL) EEPROM memory load
eml:
	rcall	skipsp
	clr	r14
	clr	r15
	ld	r16,X
	tst	r16
	breq	eml0

	rcall	rdhex0
	tst	r13
	breq	emle
	ld	r16,X
	tst	r16
	brne	emle
eml0:
	rcall	emloadm
	brcs	emle
	lds	r16,emadr+1
	add	r16,r15
	sts	emadr+1,r16
	lds	r16,emadr
	adc	r16,r14
	sts	emadr,r16

	rcall	emloadb
	brcs	emle
	rjmp	wstart
	
emle:	
	rjmp	err

emloadm:			; Load Meta Data
	ldi	XH,high(embuf)
	ldi	XL,low(embuf)
	ldi	YH,high(128)
	ldi	YL,low(128)
	ldi	r17,16
emlm0:
	rcall	ee1read
	brcs	emlme
	st	X+,r8
	adiw	YL,1
	dec	r17
	brne	emlm0

	ldi	XH,high(embuf)
	ldi	XL,low(embuf)
	ld	r16,X+
	cpi	r16,0xFF
	brne	emlme
	ld	r16,X+
	cpi	r16,0xAA
	brne	emlme
	ld	r16,X+
	cpi	r16,0x55
	brne	emlme
	ld	r16,X+
	cpi	r16,0x53
	brne	emlme
	
	clc
	ret
emlme:
	sec
	ret
	
emloadb:			; Load Binary Data
	lds	XH,emadr
	lds	XL,emadr+1
	ldi	YH,high(256)
	ldi	YL,low(256)
	lds	r19,emlen
	lds	r20,emlen+1
emlb0:
	rcall	ee1read
	brcs	emlbe
	push	YH
	push	YL
	mov	r10,XH
	mov	r11,XL
	rcall	write
	pop	YL
	pop	YH
	adiw	YL,1
	adiw	XL,1
	sec
	sbc	r20,r1
	sbc	r19,r1
	mov	r16,r20
	or	r16,r19
	brne	emlb0
	clc
	ret
emlbe:
	sec
	ret

	;; Detect memories
mdetect:
	clr	r16
	sts	mavail,r16
	ldi	r17,4
	ldi	r18,0b00000001	; (1<<PB_CS0)
mdt0:
	in	r16,PORTB
	com	r16
	or	r16,r18
	com	r16
	out	PORTB,r16

	clr	r10
	clr	r11
	rcall	read0
	mov	r19,r8		; Save

	ldi	r20,0xAA
	mov	r8,r20
	rcall	write0		; Write 0xAA

	rcall	read0
	cp	r8,r20
	brne	mdt1

	ldi	r20,0x55
	mov	r8,r20
	rcall	write0		; Write 0x55

	rcall	read0
	cp	r8,r20
	brne	mdt1

	lds	r20,mavail
	or	r20,r18
	sts	mavail,r20
mdt1:
	mov	r8,r19
	rcall	write0		; Restore

	in	r16,PORTB
	ori	r16,0b00001111
	out	PORTB,r16
	
	lsl	r18
	dec	r17
	brne	mdt0

	ret

mddisp:
	lds	r16,mavail
	andi	r16,0b00001111
	cpi	r16,0b00001111
	brne	mddsp0
	ldi	ZH,high(md32msg)
	ldi	ZL,low(md32msg)
	rjmp	mddspr
mddsp0:	
	lds	r16,mavail
	andi	r16,0b00000011
	cpi	r16,0b00000011
	brne	mddsp1
	ldi	ZH,high(md16msg)
	ldi	ZL,low(md16msg)
	rjmp	mddspr
mddsp1:	
	lds	r16,mavail
	andi	r16,0b00000001
	cpi	r16,0b00000001
	brne	mddsp2
	ldi	ZH,high(md8msg)
	ldi	ZL,low(md8msg)
	rjmp	mddspr
mddsp2:
	ldi	ZH,high(md0msg)
	ldi	ZL,low(md0msg)
mddspr:
	rcall	msgout
	rjmp	crlf

;;;
;;;
;;;
	
msgout:
        lsl     ZL
        rol     ZH
mo0:
        lpm
        adiw    ZL,1
        tst     r0
        breq    mo1
        mov     r16,r0
        rcall   conout
        rjmp    mo0
mo1:
        ret

strout:
        ld      r16,Z+
        tst     r16
        breq    so0
        rcall   conout
        rjmp    strout
so0:
        ret

hexout2:
        mov     r18,r16
        swap    r16
        rcall   hexout1
        mov     r16,r18
hexout1:
        andi    r16,0b00001111
        ori     r16,'0'
        cpi     r16,'9'+1
        brcs    ho1
        ldi     r17,'A'-'9'-1
        add     r16,r17
ho1:
        rjmp    conout

hexin:
	clr	r8
	clr	r9
	rcall	hi0
	tst	r9
	breq	hir	; Non HEX character found
	swap	r8
hi0:
	rcall	conin
	rcall	upper
	cpi	r16,'0'
	brcs	hir
	cpi	r16,'9'+1
	brcs	hi1
	cpi	r16,'A'
	brcs	hir
	cpi	r16,'F'+1
	brcc	hir
	subi	r16,'A'-'9'-1
hi1:
	subi	r16,'0'
	or	r8,r16
	inc	r9
hir:
	ret

crlf:
        ldi     r16,CR
        rcall   conout
        ldi     r16,LF
        rcall   conout
	ret

	;; Get line
getlin:
        clr     r16
        sts     llen,r16
        ldi     XL,low(inbuf)
        ldi     XH,high(inbuf)
gl0:
        rcall   conin
        andi    r16,0b01111111

        cpi     r16,BS
        breq    gl1
        cpi     r16,DEL
        brne    gl2
gl1:
        lds     r16,llen
        tst     r16
        breq    gl0
        dec     r16
        sts     llen,r16
        ld      r16,-X
        ldi     r16,BS
        rcall   conout
        ldi     r16,' '
        rcall   conout
        ldi     r16,BS
        rcall   conout
        rjmp    gl0
gl2:
        cpi     r16,CR
        breq    gl3
        cpi     r16,LF
        brne    gl4
gl3:
        rcall   crlf
        rjmp    gl9
gl4:
        cpi     r16,' '
        brcs    gl0

	;; Insert char.
	lds	r17,llen
        cpi     r17,(BUFLEN-1)
        brcc    gl0
        inc     r17
        sts     llen,r17
        st      X+,r16
        rcall   conout
        rjmp    gl0
gl9:
        clr     r16
        st      X,r16           ; Terminate
        ret

skipsp:
	ld	r16,X+
	cpi	r16,' '
	breq	skipsp
	cpi	r16,0x09
	breq	skipsp
	sbiw	XL,1
	ret

upper:
	cpi	r16,'a'
	brcs	upr
	cpi	r16,'z'+1
	brcc	upr
	subi	r16,'a'-'A'
upr:
	ret

rdhex0:
	clr	r12
rdhex:
	clr	r13
	clr	r14		; Value(H)
	clr	r15		; Value(L)
rh0:
	ld	r16,X
	rcall	upper
	cpi	r16,'0'
	brcs	rhe
	cpi	r16,'9'+1
	brcs	rh1
	cpi	r16,'A'
	brcs	rhe
	cpi	r16,'F'+1
	brcc	rhe
	subi	r16,'A'-'9'-1
rh1:
	subi	r16,'0'
	lsl	r15
	rol	r14
	lsl	r15
	rol	r14
	lsl	r15
	rol	r14
	lsl	r15
	rol	r14
	or	r15,r16	
	ld	r16,X+		; inc X
	inc	r13
	dec	r12
	brne	rh0
rhe:
	ret

	;;
	;; DPSRAM Access
	;;

read:
	if DEBUG
	lds	r16,mmode
	cpi	r16,0xFF
	brne	rd0
	mov	YH,r10
	mov	YL,r11
	ld	r8,Y
	ret
rd0:	
	endif
	
	cli

	push	r10
	push	r11

	rcall	atrans
	rcall	read0
	rcall	negcs

	pop	r11
	pop	r10
	
	sei
	ret

write:
	if DEBUG
	lds	r16,mmode
	cpi	r16,0xFF
	brne	wr0
	mov	YH,r10
	mov	YL,r11
	st	Y,r8
	ret
wr0:
	endif
	
	cli

	push	r10
	push	r11

	rcall	atrans
	rcall	write0
	rcall	negcs

	pop	r11
	pop	r10
	
	sei
	ret

atrans:
	lds	r16,mmode
	tst	r16
	brne	at1

	;; MMode: 00
	cbi	PORTB,PB_CS0	; CS0=0
	ldi	r16,0b00001111
	and	r10,r16
	ret
at1:
	cpi	r16,0x01
	brne	at2

	;; MMode: 01
	mov	r16,r11
	andi	r16,0b00000001
	brne	at10
	cbi	PORTB,PB_CS0	; CS0=0
	rjmp	at11
at10:
	cbi	PORTB,PB_CS1	; CS1=0
at11:
	clc
	ror	r10		; Addr H
	ror	r11		; Addr L
	ret
at2:

	;; MMode: else
	mov	r16,r10		; Addr H
	andi	r16,0b00010000
	brne	at20
	cbi	PORTB,PB_CS0	; CS0=0
	rjmp	at21
at20:
	cbi	PORTB,PB_CS1	; CS1=0
at21:
	ldi	r16,0b00001111
	and	r10,r16
	ret

negcs:
	in	r16,PORTB
	ori	r16,0b00001111
	out	PORTB,r16	; CS0,CS1,CS2,CS3=1
	ret

read0:	
	ldi	r16,0b11111111
	out	DDRA,r16

	out	PORTA,r10	; Addr H
	cbi	PORTD,PD_ALEH	; ALEH=0
	sbi	PORTD,PD_ALEH	; ALEH=1

	out	PORTA,r11	; Addr L
	cbi	PORTD,PD_ALEL	; ALEL=0
	sbi	PORTD,PD_ALEL	; ALEL=1

	ldi	r16,0b11111111
	out	PORTA,r16
	clr	r16
	out	DDRA,r16

	cbi	PORTD,PD_RD	; RD=0
	nop
	nop
	in	r8,PINA
	sbi	PORTD,PD_RD	; RD=1
	
	ret
	
write0:	
	ldi	r16,0b11111111
	out	DDRA,r16

	out	PORTA,r10	; Addr H
	cbi	PORTD,PD_ALEH	; ALEH=0
	sbi	PORTD,PD_ALEH	; ALEH=1

	out	PORTA,r11	; Addr L
	cbi	PORTD,PD_ALEL	; ALEL=0
	sbi	PORTD,PD_ALEL	; ALEL=1

	out	PORTA,r8

	cbi	PORTD,PD_WR	; WR=0
	nop
	nop
	sbi	PORTD,PD_WR	; WR=1

	nop
	ldi	r16,0b11111111
	out	PORTA,r16
	clr	r16
	out	DDRA,r16

	ret

	;;
	;; EEPROM routines
	;;
	;;	Addr: (YH:YL)
	;;	Data: (r8)
	;;

	;; Read BYTE from EEPROM
eeread:
	lds	r16,eenum
	tst	r16
	brne	eer0
	rjmp	ee0read
eer0:
	rjmp	ee1read

	;; Write BYTE to EEPROM
eewrite:
	lds	r16,eenum
	tst	r16
	brne	eew0
	rjmp	ee0write
eew0:
	rjmp	ee1write
	
	;; Read BYTE from internal EEPROM
ee0read:
	cli
ee0r0:	
	sbic	EECR,EEPE
	rjmp	ee0r0
	out	EEARH,YH
	out	EEARL,YL
	sbi	EECR,EERE
	in	r8,EEDR
	clc

	sei
	ret

	;; Write BYTE to internal EEPROM
ee0write:
	cli
ee0w0:	
	sbic	EECR,EEPE
	rjmp	ee0w0
	out	EEARH,YH
	out	EEARL,YL
	out	EEDR,r8
	sbi	EECR,EEMPE
	sbi	EECR,EEPE
	clc

	sei
	ret

	;; Read BYTE from target EEPROM
ee1read:
	ldi	r16,0b10100000	; 0xA0_W
	rcall	twist		; START
	brcs	ee1re
	mov	r16,YH
	rcall	twitr		; Addr H
	brcs	ee1re
	mov	r16,YL
	rcall	twitr		; Addr L
	brcs	ee1re

	ldi	r16,0b10100001	; 0xA0_R
	rcall	twist		; START
	brcs	ee1re
	rcall	twirc		; Data
	brcs	ee1re
	mov	r8,r16
	rcall	twisp		; STOP

	clc
	ret
ee1re:	
	sec
	ret

	;; Write BYTE to target EEPROM
ee1write:
	ldi	r16,0b10100000	; 0xA0_W
	rcall	twist		; START
	brcs	ee1we
	mov	r16,YH
	rcall	twitr		; Addr H
	brcs	ee1we
	mov	r16,YL
	rcall	twitr		; Addr L
	brcs	ee1we
	mov	r16,r8
	rcall	twitr		; Data
	brcs	ee1we
	rcall	twisp		; STOP

	clc
	ret
ee1we:	
	sec
	ret

	;; I2C: Transmit START & SLA
twist:
	ldi	r18,(1<<TWINT)|(1<<TWSTA)|(1<<TWEN)
	sts	TWCR,r18
tws0:	
	lds	r18,TWCR
	sbrs	r18,TWINT
	rjmp	tws0

	lds	r18,TWSR
	andi	r18,0xF8
	cpi	r18,0x08
	breq	tws1
	cpi	r18,0x10
	breq	tws1
	rjmp	twse
tws1:	
	sts	TWDR,r16	; SLA_W / SLA_R
	ldi	r18,(1<<TWINT)|(1<<TWEN)
	sts	TWCR,r18
tws2:
	lds	r18,TWCR
	sbrs	r18,TWINT
	rjmp	tws2

	lds	r18,TWSR
	andi	r18,0xF8
	cpi	r18,0x18	; SLA_W
	breq	twsf
	cpi	r18,0x40	; SLA_R
	breq	twsf
	cpi	r18,0x20	; SLA_W Not_ACK
	breq	twsrt
	cpi	r18,0x48	; SLA_R Not_ACK
	breq	twsrt
twse:
	sec
	ret
twsrt:
	rcall	twisp
	rjmp	twist
twsf:	
	clc
	ret
	
	;; I2C: Transmit DATA
twitr:
	sts	TWDR,r16	; DATA
	ldi	r18,(1<<TWINT)|(1<<TWEN)
	sts	TWCR,r18
twt0:
	lds	r18,TWCR
	sbrs	r18,TWINT
	rjmp	twt0

	lds	r18,TWSR
	andi	r18,0xF8
	cpi	r18,0x28
	brne	twte

	clc
	ret
twte:
	sec
	ret

	;; I2C: Receive DATA with ACK
twirca:
	ldi	r18,(1<<TWINT)|(1<<TWEN)|(1<<TWEA)
	sts	TWCR,r18
	rjmp	twr0
	
	;; I2C: Receive DATA
twirc:
	ldi	r18,(1<<TWINT)|(1<<TWEN)
	sts	TWCR,r18
twr0:
	lds	r18,TWCR
	sbrs	r18,TWINT
	rjmp	twr0

	lds	r18,TWSR
	andi	r18,0xF8
	cpi	r18,0x50
	breq	twrr
	cpi	r18,0x58
	breq	twrr
	
	sec
	ret
twrr:
	lds	r16,TWDR

	clc
	ret

	;; I2C: Transmit STOP
twisp:
	ldi	r18,(1<<TWINT)|(1<<TWEN)|(1<<TWSTO)
	sts	TWCR,r18
twp0:
	lds	r18,TWCR
	sbrc	r18,TWSTO
	rjmp	twp0
	;; rcall	hexout2
	;; rcall	crlf
	
	clc
	ret

	;;
	;; Interrupt handler (T0)
	;;
t0_int:
	in	r7,SREG

	;; Timer
	lds	r6,timer
	tst	r6
	breq	t0_i0

	dec	r6
	sts	timer,r6
t0_i0:	

	lds	r22,shmadr
	lds	r23,shmadr+1
	andi	r23,0b11110011	; [+0] Signature
	rcall	iread
	cpi	r21,EG_SIG
	breq	t0_i00
t0_i0r
	rjmp	t0_ir
t0_i00:	
	rcall	iainc		; [+1] Handshake
	rcall	iread
	cpi	r21,EH_REQ
	brne	t0_i0r

	rcall	iainc		; [+2] Command
	rcall	iread
	cpi	r21,EC_INI
	brne	t0_i1

	;; 00 Init (dummy)
	clr	r21
	lds	r21,UDR1	; Dummy read 
	lds	r21,UDR1	; Dummy read 
	lds	r21,UDR1	; Dummy read 
	lds	r21,UDR1	; Dummy read 
	rjmp	t0_if

t0_i1:
	cpi	r21,EC_COT
	brne	t0_i2

	;; 01 Console Output
	lds	r21,UCSR1A
	andi	r21,0b00100000
	breq	t0_ir		; Busy => do nothing

	rcall	iainc
	rcall	iainc
	rcall	iread
	sts	UDR1,r21
	clr	r21		; Status = 0
	rjmp	t0_if
	
t0_i2:
	cpi	r21,EC_CIN
	brne	t0_i3

	;; 02 Console Input
	lds	r21,UCSR1A
	andi	r21,0b10000000
	breq	t0_ir

	lds	r21,UDR1
	rcall	iainc
	rcall	iainc
	rcall	iwrite
	clr	r21
	rjmp	t0_if
	
t0_i3:
	cpi	r21,EC_CST
	brne	t0_i4

	;; 03 Console Status
	lds	r21,UCSR1A
	andi	r21,0b10000000
	rol	r21
	rol	r21
	
	rcall	iainc
	rcall	iainc
	rcall	iwrite
	clr	r21
	rjmp	t0_if
	
t0_i4:

	ldi	r21,ES_UK	; Error	

t0_if:
	lds	r22,shmadr
	lds	r23,shmadr+1
	andi	r23,0b11110011
	rcall	iainc
	rcall	iainc
	rcall	iainc		; [+3] Status
	rcall	iwrite

	rcall	iadec
	rcall	iadec		; [+1]
	ldi	r21,EH_ACK
	rcall	iwrite

t0_ir:
	out	SREG,r7
	reti

iread:
	push	r8
	push	r10
	push	r11
	push	r16

	mov	r10,r22
	mov	r11,r23
	rcall	atrans
	rcall	read0
	rcall	negcs
	mov	r21,r8
	
	pop	r16
	pop	r11
	pop	r10
	pop	r8

	ret

iwrite:
	push	r8
	push	r10
	push	r11
	push	r16

	mov	r10,r22
	mov	r11,r23
	mov	r8,r21
	rcall	atrans
	rcall	write0
	rcall	negcs

	pop	r16
	pop	r11
	pop	r10
	pop	r8

	ret
	
iainc:
	lds	r25,shmoff
	andi	r25,0b00000011
	sec
	adc	r23,r25
	adc	r22,r1
	
	ret
	
iadec:
	lds	r25,shmoff
	andi	r25,0b00000011
	sec
	sbc	r23,r25
	sbc	r22,r1

	ret
	
	;;
	;;
	;;
	
opnmsg:
        data	"\r\nEMILY Board \0"
md0msg:
	data	"\0"
md8msg:
	data	"8bit\0"
md16msg:
	data	"16bit (with +8)\0"
md32msg:
	data	"32bit (with +8 and +16)\0"
ihemsg:
        data	"Error ihex\r\n\0"
shemsg:
        data	"Error srec\r\n\0"
errmsg:
        data	"Error\r\n\0"
namsg:
        data	"N/A\r\n\0"
dsep0:
        data	" :\0"
dsep1:
        data	" : \0"
ihexer:
        data	":00000001FF\r\n\0"
srecer:
        data	"S9030000FC\r\n\0"
shmmsg0:
        data	"SHM Address: \0"
shmmsg1:
        data	"- (+\0"
mmdmsg:
	data	"Memory Mode: \0"
ldrmsg:
	data	"Loaded: \0"

	if USE_DEV_MEGA164
	include	"dev/mega164.asm"
	endif

	;;
	;; Work Area
	;;
	
	segment	data

	org	WORK_B
	
inbuf:	res	BUFLEN		; Line input buffer
dsadrh: res	1       	; DUMP start address (H)
dsadrl: res	1		; DUMP start address (L)
deadrh: res	1       	; DUMP end address (H)
deadrl: res	1       	; DUMP end address (L)
dstate: res	1
saddrh: res	1       	; SET address (H)
saddrl: res	1		; SET address (L)
llen:	res	1
lsaddr:	res	2		; LOAD start address
leaddr:	res	2		; LOAD end address
eenum:	res	1		; EEPROM number 0:mega164,1:target
mavail:	res	1		; Available memory

timer:	res	1		; Timer

	;; EEPROM Buffer
epbuf:
eesig0:	res	4		; Signature FF:AA:55:50

hexmod:	res	1		; HEX file mode
	
shmadr:	res	2		; Shared Memory Address
shmoff:	res	1		; Shared Memory Offset

mmode:	res	1		; Memory mode

eppad:	res	(epbuf+32-eppad) ; Padding
epend:	

eptmp:	res	32


embuf:
emsig0:	res	4		; Signature FF:AA:55:53

emadr:	res	2		; Address
emlen:	res	2		; Length

empad:	res	(embuf+16-empad) ; Padding
emend:

emtmp:	res	16

	end
